/* ************************************************************************** */
/*                                                                            */
/*                                                        :::      ::::::::   */
/*   manage_threads.c                                   :+:      :+:    :+:   */
/*                                                    +:+ +:+         +:+     */
/*   By: erey-bet <marvin@42.fr>                    +#+  +:+       +#+        */
/*                                                +#+#+#+#+#+   +#+           */
/*   Created: 2023/03/29 20:44:30 by erey-bet          #+#    #+#             */
/*   Updated: 2023/04/10 15:55:42 by erey-bet         ###   ########.fr       */
/*                                                                            */
/* ************************************************************************** */

#include "philo.h"

static t_same	*create_struct_same(t_config *config)
{
	t_same	*same;

	same = ft_calloc(sizeof(t_same), 1);
	if (!same)
		return (NULL);
	same->forks = init_forks(config);
	if (!same->forks)
		return (same);
	same->mutex = ft_calloc(sizeof(pthread_mutex_t), config->nbr_philo + 3);
	if (!same->mutex)
		return (same);
	if (init_mutex(same->mutex, config->nbr_philo + 3))
		return (same);
	same->config = config;
	same->death = 0;
	same->all_eat = 0;
	same->time = get_utime();
	return (same);
}

static t_only	*create_struct_only(int id)
{
	t_only	*only;

	only = ft_calloc(sizeof(t_only), 1);
	if (!only)
		return (NULL);
	only->id = id;
	only->eat = 0;
	return (only);
}

static t_philo	*init_philo(t_same *same, t_only *only)
{
	t_philo	*philo;

	if (!same)
	{
		if (only)
			free(only);
		return (NULL);
	}
	if (!only)
	{
		free(same);
		return (NULL);
	}
	philo = ft_calloc(sizeof(t_philo), 1);
	if (!philo)
		return (NULL);
	philo->same = same;
	philo->only = only;
	return (philo);
}

int	manage_threads(t_config *config)
{
	pthread_t	*thds;
	t_philo		**philo;
	t_same		*same;
	int			i;

	thds = ft_calloc(config->nbr_philo, sizeof(pthread_t));
	if (thds == NULL)
		return (1);
	philo = ft_calloc(config->nbr_philo + 1, sizeof(t_philo *));
	if (philo == NULL)
		return (free_all(thds, NULL, NULL));
	same = create_struct_same(config);
	i = -1;
	while (++i < config->nbr_philo)
	{
		philo[i] = init_philo(same, create_struct_only(i));
		if (!philo[i] || pthread_create(&thds[i], NULL,
				(void *)&philosopher, philo[i]) != 0)
			return (free_all(thds, same, philo));
	}
	philo[i] = NULL;
	while (--i > -1)
		pthread_join(thds[i], NULL);
	free_all(thds, same, philo);
	return (0);
}
